/*
** -----------------------------------------------------------------------------**
** stay_alive.v
**
** Manages power, turns off when not in use
**
** Copyright (C) 2003-2006 Elphel, Inc.
**
** -----------------------------------------------------------------------------**
**  This file is part of x324
**  x324 is free software - hardware description language (HDL) code.
** 
**  This program is free software: you can redistribute it and/or modify
**  it under the terms of the GNU General Public License as published by
**  the Free Software Foundation, either version 3 of the License, or
**  (at your option) any later version.
**
**  This program is distributed in the hope that it will be useful,
**  but WITHOUT ANY WARRANTY; without even the implied warranty of
**  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**  GNU General Public License for more details.
**
**  You should have received a copy of the GNU General Public License
**  along with this program.  If not, see <http://www.gnu.org/licenses/>.
** -----------------------------------------------------------------------------**
**
**  $Log: stay_alive.v,v $
**  Revision 1.1.1.1  2008/11/27 20:04:02  elphel
**
**
**  Revision 1.1.1.1  2007/08/17 12:07:19  elphel
**  This is a fresh tree based on elphel353-2.10
**
**  Revision 1.6  2007/08/17 12:07:19  spectr_rain
**  switch to GPL3 license
**
**  Revision 1.5  2007/07/20 09:46:13  spectr_rain
**  *** empty log message ***
**
**  Revision 1.4  2007/05/25 02:09:09  elphel
**  347 update for mechanical shutter
**
**  Revision 1.1  2006/09/24 16:44:36  elphel
**  Split file into separate module files, added real time clock module, project files
**
*/
module stay_alive(clk,     //25MHz
                  we,    
                  a,       // [2:0] reg address (000 - watchdog in , 001 - seq. on delay, 010 - power off delay (0.001 sec)
                           // 4 - PWM period (in pixel clocks), 5 - PWM start (= period - longest)
                           // 6 - PWM decrement (each period)
                  
                  d,       // [15:0] (watchdog - 16 bits (65.536 sec max), others - 10 bits (1.024 sec max)
                  
                  en,      // enable (was power on)
                  trig,    // get picture cmd (1 cycle)
                  pwren,   // power enable output 
                  seqen,   // enable sequencer
                  seqtrig);// trigger sequencer
 parameter N24999=15'h61a7; // **************** change for simulation

   input         clk, we, trig, en;
   input  [ 2:0] a;
   input  [15:0] d;
   output        pwren, seqen,seqtrig;

   reg    [15:0] wdog_dur;
   reg    [ 9:0] sqon_dur;
   reg    [ 9:0] pwroff_dur;

   reg    [14:0] cntr_1ms;
   reg           p_1ms;
   reg           wdog, sqon_dly;
   wire          pwroff_dly; // needs init for simulator
   reg    [15:0] wdog_cntr;
   reg    [ 9:0] sqon_cntr;
   reg    [ 9:0] pwroff_cntr;
   reg           seqen, seqen_d;
   wire          first_seqtrig;  // delayed by >2 cycles after starting sequencer - will trigger trhe first time
   reg           seqtrig;
   reg           pwren;
   reg           wdog_no; // no watchdog if time is set to 0xffff (65535) msec
   
   reg    [15:0] pwm_period;
   reg    [15:0] pwm_start;
   reg    [15:0] pwm_end;
   reg    [15:0] pwm_decrement;
   reg    [15:0] pwm_phase; // counts pwm_period to 0
   reg    [15:0] pwm_value; // counts from pwm_start down to pwm_end with step pwm_decrement
   reg           pwm_on;    // 1 when pwm signal is on
   reg           pwm_done;  // 1 when ramp is over
   reg           pwm_tick;  // end of PWM period
   wire   [16:0] pwm_next= {1'b0,pwm_value[15:0]}-{1'b0,pwm_decrement[15:0]};
   wire   [16:0] pwm_diff= {1'b0,pwm_phase[15:0]}-{1'b0,pwm_value[15:0]};
   wire   [16:0] pwm_left= {1'b0,pwm_value[15:0]}-{1'b0,pwm_end[15:0]};
   always @ (posedge clk) begin
// write values
     if (we && (a[2:0]==3'h0)) wdog_dur[15:0]      <= d[15:0];
     if (we && (a[2:0]==3'h1)) sqon_dur[ 9:0]      <= d[ 9:0];
     if (we && (a[2:0]==3'h2)) pwroff_dur[ 9:0]    <= d[ 9:0];
     if (we && (a[2:0]==3'h4)) pwm_period[15:0]    <= d[15:0];
     if (we && (a[2:0]==3'h5)) pwm_start[15:0]     <= d[15:0];
     if (we && (a[2:0]==3'h6)) pwm_decrement[15:0] <= d[15:0];
     if (we && (a[2:0]==3'h7)) pwm_end[15:0]      <= d[15:0];
     wdog_no <= &wdog_dur[15:0];
// timing marks 
     p_1ms <= ((en || pwren) && (cntr_1ms[14:0]==15'h1));
     if (!(en || pwren) || p_1ms) cntr_1ms[14:0] <= N24999; //15'h61a7; // 24999 
     else                         cntr_1ms[14:0] <= cntr_1ms[14:0]-1;
// watchdog (should be >=2)
     if      (!en)           wdog_cntr[15:0] <= 16'h0;
     else if (trig)          wdog_cntr[15:0] <= wdog_dur[15:0];
//     else if (wdog && p_1ms) wdog_cntr[15:0] <= wdog_cntr[15:0]-1;
     else if (wdog && p_1ms && !wdog_no) wdog_cntr[15:0] <= wdog_cntr[15:0]-1;
     wdog <= en && (trig || (wdog_cntr[15:0]!=16'b0));
// power off delay  ( FD prevents it from being undefined in simulator)
     if      (!pwroff_dly)  pwroff_cntr[9:0] <= 10'b0;
     else if (seqen)        pwroff_cntr[9:0] <= pwroff_dur[9:0];
     else if (p_1ms)        pwroff_cntr[9:0] <= pwroff_cntr[9:0]-1;
     seqen <= en && (wdog || (seqen && trig)) && !sqon_dly;
     seqen_d <= seqen;
     seqtrig <= trig || first_seqtrig;
//     pwren   <= wdog || pwroff_dly;
     pwren   <= pwm_on || pwroff_dly;
     
     pwm_tick <= wdog && (pwm_phase[15:0]==16'h1);
     
     pwm_done <= wdog && (pwm_done || (pwm_tick && (pwm_next[16] || pwm_left[16]))) || (pwm_period[15:0]==16'h0);
     
     if      (!wdog)    pwm_value[15:0] <= pwm_start[15:0];
     else if (pwm_tick) pwm_value[15:0] <= pwm_next[15:0];
     
     if      (!wdog || pwm_tick)    pwm_phase[15:0] <= pwm_period[15:0];
     else if (!pwm_done)            pwm_phase[15:0] <= pwm_phase[15:0]-1;
     
     pwm_on <= wdog && (pwm_done || ~pwm_diff[16]);

// sequencer on delay
//     if      (!wdog)             sqon_cntr[9:0] <= sqon_dur[9:0];
     if      (!wdog || !pwm_done) sqon_cntr[9:0] <= sqon_dur[9:0];
     else if (sqon_dly && p_1ms)  sqon_cntr[9:0] <= sqon_cntr[9:0]-1;
     sqon_dly <= (sqon_cntr[9:0]!=10'b0);

     
   end
/*
   reg    [15:0] pwm_period;
   reg    [15:0] pwm_start;
   reg    [15:0] pwm_decrement;
   reg    [15:0] pwm_phase; // counts pwm_period to 0
   reg    [15:0] pwm_value; // counts from pwm_start down to 0 with step pwm_decrement
   reg           pwm_on;    // 1 when pwm signal is on
   reg           pwm_done;  // 1 when ramp is over

*/   
   
   
   
  FD    i_pwroff_dly    (.C(clk),   .D (seqen || (pwroff_dly && (pwroff_cntr[9:0]!=10'b0))), .Q(pwroff_dly));
  SRL16 i_first_seqtrig (.CLK(clk), .Q(first_seqtrig),.D(seqen && !seqen_d), .A0(1'b0), .A1(1'b1), .A2(1'b0), .A3(1'b0));

endmodule
