/*
** -----------------------------------------------------------------------------**
** camsync.v
**
** Synchronization between cameras using GPIO lines on th 10353 board:
**  - triggering from selected line(s) with filter;
**  - programmable delay to actual trigger (in pixel clock periods)
**  - Generating trigger output to selected GPIO line (and polarity)
**    or directly to the input delay generator (see bove)
**  - single/repetitive output with specified period in pixel clocks
**
** Copyright (C) 2007 Elphel, Inc
**
** -----------------------------------------------------------------------------**
**  This file is part of X353
**  X353 is free software - hardware description language (HDL) code.
** 
**  This program is free software: you can redistribute it and/or modify
**  it under the terms of the GNU General Public License as published by
**  the Free Software Foundation, either version 3 of the License, or
**  (at your option) any later version.
**
**  This program is distributed in the hope that it will be useful,
**  but WITHOUT ANY WARRANTY; without even the implied warranty of
**  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**  GNU General Public License for more details.
**
**  You should have received a copy of the GNU General Public License
**  along with this program.  If not, see <http://www.gnu.org/licenses/>.
** -----------------------------------------------------------------------------**
**
*/

module camsync       (sclk, // @negedge
                         pre_wen, // 1 cycle ahead of write data
//                         di,      // [31:0] data in
                         di,      // [15:0] data in
                         wa,      // [1:0] write address
                           // 0 - source of trigger (12 bit pairs, LSB - level to trigger, MSB - use this bit). All 0 - internal trigger
                           // 1 - input trigger delay (pixel clocks), 0 - trigger disabled
                           // 2 - 24 bit pairs: MSB - enable selected line, LSB - level to send when trigger active
									//     bit 25==1 some of the bits use test mode signals:
                           // 3 - output trigger period (duration constant of 256 pixel clocks). 
									//     d==0 - disable (stop periodic mode)
									//     0<d<256(?) single trigger
									//     256>=d - repetitive trigger
								 pclk,    // pixel clock (global)
                         trigrst,   // single-clock start of frame input (resets trigger output) posedge
								 gpio_in, // 12-bit input from GPIO pins
								 gpio_out,// 12-bit output to GPIO pins
								 gpio_out_en,// 12-bit output enable to GPIO pins
								 trigger1, // 1 cycle-long trigger output
								 trigger); // active high trigger to the sensor (reset by 
    input          sclk;
    input          pre_wen;
//    input  [31:0] di;
    input  [15:0] di;
    input  [ 1:0] wa;
	 input           pclk;
	 input           trigrst; //vacts;
    input  [11:0] gpio_in;
    output [11:0] gpio_out;
    output [11:0] gpio_out_en;
    output         trigger1;
    output         trigger;
// delaying everything by 1 clock to reduce data fan in
    reg    [1:0]  wad;
    reg   [31:0]  did;
    reg             pre_wend;	 
	 reg    [3:0]  wen;
	 wire            high_zero; // 24 MSBs are zero 
    reg    [11:0] input_use;       // 1 - use this bit
    reg    [11:0] input_pattern;   // data to be compared for trigger event to take place
	 reg             input_use_intern;//
	 reg    [31:0] input_dly;       // delay value for the trigger
	 reg    [11:0] gpio_out_en;     // which GPIO lines to drive
	 reg    [11:0] gpio_active;     // output levels on the selected GPIO lines during output pulse (will be negated when inactive)
	 reg           testmode;        // drive some internal signals to GPIO bits
	 reg             outsync;
	 reg    [31:0] repeat_period;    // restart period in repetitive mode
	 reg             start,start_d;   // start single/repetitive output pulse(s)
	 reg             rep_en;          // enable repetitive mode
	 reg             start_en;
	 wire            start_to_pclk;
	 reg    [2:0]  start_pclk; // start and restart
	 reg   [31:0]  restart_cntr; // restart period counter
	 reg    [1:0]  restart_cntr_run; // restart counter running
	 wire          restart;          // restart out sync
	 reg    [8:0]  out_pulse_cntr;
	 reg           trigger_condition; // GPIO input trigger condition met
	 reg           trigger_condition_d; // GPIO input trigger condition met, delayed (for edge detection)
	 reg           trigger_condition_filtered; // trigger condition filtered
	 reg           trigger_condition_filtered_d; // trigger condition filtered, delayed (for edge detection)
	 reg    [6:0]  trigger_filter_cntr;
    reg           trigger1;
//    reg           trigger;
    wire          trigger;       // for happy simulator
	 reg           start_dly;      // start delay (external input filtered or from internal single/rep)
	 reg   [31:0]  dly_cntr;       // trigger delay counter
//	 reg           dly_cntr_run;   // trigger delay counter running
	 wire          dly_cntr_run;   // trigger delay counter running (to use FD for simulation)
	 reg           dly_cntr_run_d; // trigger delay counter running - delayed by 1
	 

	 assign  high_zero = did[31:8]==24'b0;

//! in testmode GPIO[11] and GPIO[10] use internal signals instead of the outsync:
//! bit 11 - same as TRIGGER output to the sensor (signal to the sensor may be disabled externally)
//!          then that bit will be still from internall trigger to frame valid
//! bit 10 - dly_cntr_run (delay counter run) - active during trigger delay
	 assign  gpio_out[9: 0] = outsync? gpio_active[9: 0]: ~gpio_active[9: 0];
	 assign  gpio_out[10] = (testmode? dly_cntr_run: outsync)? gpio_active[10]: ~gpio_active[10];
	 assign  gpio_out[11] = (testmode? trigger:      outsync)? gpio_active[11]: ~gpio_active[11];

	 assign  restart= restart_cntr_run[1] && !restart_cntr_run[0];
    always @ (negedge sclk) begin
	   pre_wend <= pre_wen;
		if (pre_wen)  wad[1:0] <= wa[1:0];
//		if (pre_wend) begin // preserving energy
//		  did[31:0] <= di[31:0];
//		end
		if (pre_wen)  did[15: 0] <= di[15:0];
		if (pre_wend) did[31:16] <= di[15:0];

		wen[3:0] <= pre_wend?{(wad[1:0]==2'h3),(wad[1:0]==2'h2),(wad[1:0]==2'h1),(wad[1:0]==2'h0)}:4'b0;
		if (wen[0]) input_use[11:0] <= {did[23],did[21],did[19],did[17],did[15],did[13],did[11],did[9],did[7],did[5],did[3],did[1]};
		input_use_intern <= (input_use[11:0]==12'h0); // use internal source for triggering
		if (wen[0]) input_pattern[11:0] <= {did[22],did[20],did[18],did[16],did[14],did[12],did[10],did[8],did[6],did[4],did[2],did[0]};
		if (wen[1]) input_dly[31:0] <= did[31:0];
		if (wen[2]) gpio_out_en[11:0] <= {did[23],did[21],did[19],did[17],did[15],did[13],did[11],did[9],did[7],did[5],did[3],did[1]};
		if (wen[2]) gpio_active[11:0] <= {did[22],did[20],did[18],did[16],did[14],did[12],did[10],did[8],did[6],did[4],did[2],did[0]};
		if (wen[2]) testmode <= did[24];
		if (wen[3]) repeat_period[31:0] <= did[31:0];

      start <= wen[3] && (did[31:0]!=32'h0);
		start_d <= start;

		start_en <= (repeat_period[31:0]!=0);
		if (wen[3]) rep_en <= !high_zero;
    end
//! synchronize start from sclk to pclk	 
//! Generating repetition (period should be exactly N, not N+/-1) and output pulse
	 FDCE_1 i_start_to_pclk (.C(sclk), .CE(start_d),.CLR(start_pclk[1] || !start_en),.D(1'b1),.Q(start_to_pclk));
	 always @ (posedge pclk) begin
//	   start_pclk[1:0] <= {start_pclk[0] || (restart && rep_en), start_to_pclk && !start_pclk[0]};
	   start_pclk[2:0] <= {(restart && rep_en) || (start_pclk[1] && !restart_cntr_run[1] && !restart_cntr_run[0] && !start_pclk[2]),
                          start_pclk[0],
                          start_to_pclk && !start_pclk[0]};
		
//	   start_pclk[1:0] <= {start_pclk[0] || (restart && rep_en), start_to_pclk && !start_pclk[0]};
 //   now start_pclk[1] - single pclk start output 
      restart_cntr_run[1:0] <= {restart_cntr_run[0],start_en && (start_pclk[2] || (restart_cntr_run[0] && (restart_cntr[31:2] !=0)))};
		if (restart_cntr_run[0]) restart_cntr[31:0] <= restart_cntr[31:0] - 1;
		else restart_cntr[31:0] <= repeat_period[31:0];
		outsync <= start_en && (start_pclk[2] || (outsync && !out_pulse_cntr[8]));
      if (start_pclk[2] || outsync) out_pulse_cntr[8:0] <= out_pulse_cntr[8:0]+1;
		else out_pulse_cntr[8:0] <= 9'h0;
	 end
 
// Detecting input sync pulse (filter - 64 pclk, pulse is 256 pclk)
	 FD i_dly_cntr_run (.C(pclk),.D(start_dly || (dly_cntr_run && (dly_cntr[31:0]!=0))),.Q(dly_cntr_run)); // for simulator to be happy
	 FD i_trigger      (.C(pclk),.D(trigger1  || (trigger && !trigrst)),  .Q(trigger)); // for simulator to be happy
	 always @ (posedge pclk) begin
	   trigger_condition <= (((gpio_in[11:0] ^ input_pattern[11:0]) & input_use[11:0]) == 12'b0);
		trigger_condition_d <= trigger_condition;
		
		if (trigger_condition !=trigger_condition_d) trigger_filter_cntr <= 0;
		else if (!trigger_filter_cntr[6]) trigger_filter_cntr<=trigger_filter_cntr+1;
		
		if (input_use_intern) trigger_condition_filtered <= 1'b0;
		else if (trigger_filter_cntr[6]) trigger_condition_filtered <= trigger_condition_d;
		
		trigger_condition_filtered_d <=  trigger_condition_filtered;

      start_dly <= input_use_intern ? (start_pclk[2] && start_en) : (trigger_condition_filtered && !trigger_condition_filtered_d);
// simulation problems w/o "start_en &&" ?		
//      dly_cntr_run   <= start_en && ( start_dly || (dly_cntr_run && (dly_cntr[31:0]!=0)));
//      dly_cntr_run   <= start_dly || (dly_cntr_run && (dly_cntr[31:0]!=0));
      dly_cntr_run_d <= dly_cntr_run;
		if (dly_cntr_run) dly_cntr[31:0] <= dly_cntr[31:0] -1;
		else              dly_cntr[31:0] <= input_dly[31:0];
		trigger1 <= dly_cntr_run_d && !dly_cntr_run;
//		trigger  <= start_en && (trigger1  || (trigger && !trigrst));
    end	 
endmodule
